@@ -725,9 +725,7 @@ static void mark_damaged(PixmapPtr pixmap, struct sna_pixmap *priv,
 	    box->x2 >= pixmap->drawable.width &&
 	    box->y2 >= pixmap->drawable.height) {
 		sna_damage_destroy(&priv->cpu_damage);
-		sna_damage_all(&priv->gpu_damage,
-			       pixmap->drawable.width,
-			       pixmap->drawable.height);
+		sna_damage_all(&priv->gpu_damage, pixmap);
 		list_del(&priv->flush_list);
 	} else {
 		sna_damage_add_box(&priv->gpu_damage, box);
@@ -791,7 +789,8 @@ trap_upload(PicturePtr picture,
 		return true;
 
 	memset(scratch->devPrivate.ptr, 0, scratch->devKind*height);
-	image = pixman_image_create_bits(picture->format, width, height,
+	image = pixman_image_create_bits((pixman_format_code_t)picture->format,
+					 width, height,
 					 scratch->devPrivate.ptr,
 					 scratch->devKind);
 	if (image) {
